"""
This module exposes the AsyncTextFileContentsManager that allows to open
text files as notebooks
"""
import inspect
import itertools
import os

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from collections import namedtuple
from datetime import timedelta

import nbformat
from tornado.web import HTTPError

# import notebook.transutils before notebook.services.contents.filemanager #75
try:
    import notebook.transutils  # noqa
except ImportError:
    pass

from .async_pairs import read_pair, write_pair
from .config import (
    JUPYTEXT_CONFIG_FILES,
    PYPROJECT_FILE,
    JupytextConfiguration,
    JupytextConfigurationError,
    find_global_jupytext_configuration_file,
    load_jupytext_configuration_file,
    notebook_formats,
    preferred_format,
)
from .formats import (
    long_form_multiple_formats,
    short_form_multiple_formats,
    short_form_one_format,
)
from .jupytext import drop_text_representation_metadata, reads, writes
from .kernels import set_kernelspec_from_language
from .paired_paths import (
    InconsistentPath,
    base_path,
    find_base_path_and_format,
    full_path,
    paired_paths,
)
from .pairs import PairedFilesDiffer, latest_inputs_and_outputs


def build_async_jupytext_contents_manager_class(base_contents_manager_class):
    """
    Derives an asynchronous TextFileContentsManager class from the given base class,
    which is supposed to be asynchronous too.
    """

    class AsyncJupytextContentsManager(
        base_contents_manager_class, JupytextConfiguration
    ):
        """
        A FileContentsManager Class that reads and stores notebooks to classical
        Jupyter notebooks (.ipynb), R Markdown notebooks (.Rmd), Julia (.jl),
        Python (.py) or R scripts (.R)
        """

        def __init__(self, *args, **kwargs):
            # Dictionary: notebook path => (fmt, formats) where
            # fmt is the current format, and formats the paired formats.
            self.paired_notebooks = dict()

            # Configuration cache, useful when notebooks are listed in a given directory
            self.cached_config = namedtuple("cached_config", "path config_file config")
            self.super = super()
            self.super.__init__(*args, **kwargs)

        def all_nb_extensions(self, config):
            """All extensions that should be classified as notebooks"""
            return [
                ext if ext.startswith(".") else "." + ext
                for ext in config.notebook_extensions
            ]

        def drop_paired_notebook(self, path):
            """Remove the current notebook from the list of paired notebooks"""
            if path not in self.paired_notebooks:
                return

            fmt, formats = self.paired_notebooks.pop(path)
            prev_paired_paths = paired_paths(path, fmt, formats)
            for alt_path, _ in prev_paired_paths:
                if alt_path in self.paired_notebooks:
                    self.drop_paired_notebook(alt_path)

        def update_paired_notebooks(self, path, formats):
            """Update the list of paired notebooks to include/update the current pair"""
            if not formats:
                self.drop_paired_notebook(path)
                return

            formats = long_form_multiple_formats(formats)
            _, fmt = find_base_path_and_format(path, formats)
            new_paired_paths = paired_paths(path, fmt, formats)
            for alt_path, _ in new_paired_paths:
                self.drop_paired_notebook(alt_path)

            if len(formats) == 1 and set(formats[0]) <= {"extension"}:
                return

            short_formats = short_form_multiple_formats(formats)
            for alt_path, alt_fmt in new_paired_paths:
                self.paired_notebooks[alt_path] = (
                    short_form_one_format(alt_fmt),
                    short_formats,
                )

        async def create_prefix_dir(self, path, fmt):
            """Create the prefix dir, if missing"""
            if "prefix" in fmt and "/" in path:
                parent_dir = self.get_parent_dir(path)
                if not await self.dir_exists(parent_dir):
                    await self.create_prefix_dir(parent_dir, fmt)
                    self.log.info("Creating directory %s", parent_dir)
                    await self.super.save(dict(type="directory"), parent_dir)

        async def save(self, model, path=""):
            """Save the file model and return the model with no content."""
            if model["type"] != "notebook":
                return await self.super.save(model, path)

            path = path.strip("/")
            nbk = model["content"]
            try:
                config = await self.get_config(path)
                jupytext_formats = notebook_formats(nbk, config, path)
                self.update_paired_notebooks(path, jupytext_formats)

                async def save_one_file(path, fmt):
                    if "format_name" in fmt and fmt["extension"] not in [
                        ".md",
                        ".markdown",
                        ".Rmd",
                    ]:
                        self.log.info(
                            "Saving %s in format %s:%s",
                            os.path.basename(path),
                            fmt["extension"][1:],
                            fmt["format_name"],
                        )
                    else:
                        self.log.info("Saving %s", os.path.basename(path))

                    await self.create_prefix_dir(path, fmt)
                    if fmt["extension"] == ".ipynb":
                        return await self.super.save(
                            dict(
                                type="notebook",
                                content=drop_text_representation_metadata(
                                    model["content"]
                                ),
                            ),
                            path,
                        )

                    if (
                        model["content"]["metadata"]
                        .get("jupytext", {})
                        .get("notebook_metadata_filter")
                        == "-all"
                    ):
                        self.log.warning(
                            "Stripping metadata from {} as 'Include Metadata' is off "
                            "(toggle 'Include Metadata' in the Jupytext Menu or Commands if desired)".format(
                                path
                            )
                        )

                    text_model = dict(
                        type="file",
                        format="text",
                        content=writes(
                            nbformat.from_dict(model["content"]), fmt=fmt, config=config
                        ),
                    )

                    return await self.super.save(text_model, path)

                return await write_pair(path, jupytext_formats, save_one_file)

            except Exception as e:
                self.log.error("Error while saving file: %s %s", path, e, exc_info=True)
                raise HTTPError(500, f"Unexpected error while saving file: {path} {e}")

        async def _get_with_no_require_hash_argument(
            self,
            path,
            content=True,
            type=None,
            format=None,
            load_alternative_format=True,
        ):
            return await self._get_with_require_hash_argument(
                path,
                content=content,
                type=type,
                format=format,
                require_hash=False,
                load_alternative_format=load_alternative_format,
            )

        async def _get_with_require_hash_argument(
            self,
            path,
            content=True,
            type=None,
            format=None,
            require_hash=False,
            load_alternative_format=True,
        ):
            """Takes a path for an entity and returns its model"""
            path = path.strip("/")
            ext = os.path.splitext(path)[1]

            super_kwargs = {"content": content, "type": type, "format": format}
            if require_hash:
                super_kwargs["require_hash"] = require_hash

            # Not a notebook?
            if (
                not await self.file_exists(path)
                or await self.dir_exists(path)
                or (type is not None and type != "notebook")
            ):
                return await self.super.get(path, **super_kwargs)

            config = await self.get_config(path, use_cache=content is False)
            if ext not in self.all_nb_extensions(config):
                return await self.super.get(path, **super_kwargs)

            fmt = preferred_format(ext, config.preferred_jupytext_formats_read)
            if ext == ".ipynb":
                super_kwargs["type"] = "notebook"
                model = await self.super.get(path, **super_kwargs)
            else:
                super_kwargs["type"] = "file"
                super_kwargs["format"] = "text"
                model = await self.super.get(path, **super_kwargs)
                model["type"] = "notebook"
                if content:
                    # We may need to update these keys, inherited from text files formats
                    # Cf. https://github.com/mwouts/jupytext/issues/659
                    model["format"] = "json"
                    model["mimetype"] = None
                    try:
                        model["content"] = reads(
                            model["content"], fmt=fmt, config=config
                        )
                        # mark all code cells from text notebooks as 'trusted'
                        # as they don't have any outputs, cf. #941
                        for cell in model["content"].cells:
                            if cell.cell_type == "code":
                                cell["metadata"]["trusted"] = True

                    except Exception as err:
                        self.log.error(
                            "Error while reading file: %s %s", path, err, exc_info=True
                        )
                        raise HTTPError(500, str(err))

            if not load_alternative_format:
                return model

            # We will now read a second file if this is a paired notebooks.
            if content:
                nbk = model["content"]
                formats = nbk.metadata.get("jupytext", {}).get(
                    "formats"
                ) or config.default_formats(path)
                formats = long_form_multiple_formats(
                    formats, nbk.metadata, auto_ext_requires_language_info=False
                )
            else:
                if path not in self.paired_notebooks:
                    return model

                _, formats = self.paired_notebooks.get(path)
                formats = long_form_multiple_formats(formats)

            # Compute paired notebooks from formats
            alt_paths = [(path, fmt)]
            if formats:
                try:
                    _, fmt = find_base_path_and_format(path, formats)
                    alt_paths = paired_paths(path, fmt, formats)
                    self.update_paired_notebooks(path, formats)
                except InconsistentPath as err:
                    self.log.error(
                        "Unable to read paired notebook: %s %s",
                        path,
                        err,
                        exc_info=True,
                    )
                    raise HTTPError(
                        500, f"Unable to read paired notebook: {path} {err}"
                    )
            else:
                if path in self.paired_notebooks:
                    fmt, formats = self.paired_notebooks.get(path)
                    alt_paths = paired_paths(path, fmt, formats)
                    formats = long_form_multiple_formats(formats)

            if content and len(alt_paths) > 1 and ext == ".ipynb":
                # Apply default options (like saving and reloading would do)
                jupytext_metadata = model["content"]["metadata"].get("jupytext", {})
                config.set_default_format_options(jupytext_metadata, read=True)
                if jupytext_metadata:
                    model["content"]["metadata"]["jupytext"] = jupytext_metadata

            async def get_timestamp(alt_path):
                # self.exists is not async for AsyncLargeManager,
                # but MetaManager from fs_manager has an async version
                exists = self.exists(alt_path)
                if inspect.isawaitable(exists):
                    exists = await exists
                if not exists:
                    return None
                if alt_path == path:
                    return model["last_modified"]
                return (await self.super.get(alt_path, content=False))["last_modified"]

            async def read_one_file(alt_path, alt_fmt):
                if alt_path == path:
                    return model["content"]
                if alt_path.endswith(".ipynb"):
                    self.log.info(f"Reading OUTPUTS from {alt_path}")
                    return (
                        await self.super.get(
                            alt_path, content=True, type="notebook", format=format
                        )
                    )["content"]

                self.log.info(f"Reading SOURCE from {alt_path}")
                text = (
                    await self.super.get(
                        alt_path,
                        content=True,
                        type="file",
                        # Don't use the parent format, see https://github.com/mwouts/jupytext/issues/1124
                        format=None,
                    )
                )["content"]
                return reads(text, fmt=alt_fmt, config=config)

            timestamps = {
                alt_path: await get_timestamp(alt_path)
                for alt_path, alt_fmt in paired_paths(path, fmt, formats)
            }

            inputs, outputs = latest_inputs_and_outputs(
                path,
                fmt,
                formats,
                lambda alt_path: timestamps[alt_path],
                contents_manager_mode=True,
            )

            # Modification time of a paired notebook is the timestamp of inputs #118 #978
            model["last_modified"] = inputs.timestamp

            if require_hash:
                if (
                    inputs.path is not None
                    and outputs.path is not None
                    and inputs.path != outputs.path
                ):
                    model_other = await self.super.get(
                        inputs.path if path == outputs.path else outputs.path,
                        content=False,
                        require_hash=True,
                    )
                    # The hash of a paired file is the concatenation of
                    # the hashes of the input and output files
                    if path == outputs.path:
                        model["hash"] = model_other["hash"] + model["hash"]
                    else:
                        model["hash"] = model["hash"] + model_other["hash"]

            if not content:
                return model

            # Before we combine the two files, we make sure we're not overwriting ipynb cells
            # with an outdated text file
            content = None
            try:
                if (
                    outputs.timestamp
                    and outputs.timestamp
                    > inputs.timestamp
                    + timedelta(seconds=config.outdated_text_notebook_margin)
                ):
                    ts_mismatch = (
                        "{out} (last modified {out_last}) is more recent than "
                        "{src} (last modified {src_last})".format(
                            src=inputs.path,
                            src_last=inputs.timestamp,
                            out=outputs.path,
                            out_last=outputs.timestamp,
                        )
                    )
                    self.log.warning(ts_mismatch)

                    try:
                        content = await read_pair(
                            inputs, outputs, read_one_file, must_match=True
                        )
                        self.log.warning(
                            "The inputs in {src} and {out} are identical, "
                            "so the mismatch in timestamps was ignored".format(
                                src=inputs.path, out=outputs.path
                            )
                        )
                    except HTTPError:
                        raise
                    except PairedFilesDiffer as diff:
                        raise HTTPError(
                            400,
                            """{ts_mismatch}

Differences (jupytext --diff {src} {out}) are:
{diff}
Please either:
- open {src} in a text editor, make sure it is up to date, and save it,
- or delete {src} if not up to date,
- or increase check margin by adding, say,
outdated_text_notebook_margin = 5  # default is 1 (second)
to your jupytext.toml file
                        """.format(
                                ts_mismatch=ts_mismatch,
                                src=inputs.path,
                                out=outputs.path,
                                diff=diff,
                            ),
                        )
            except OverflowError:
                pass

            if content is not None:
                model["content"] = content
            else:
                try:
                    model["content"] = await read_pair(inputs, outputs, read_one_file)
                except HTTPError:
                    raise
                except Exception as err:
                    self.log.error(
                        "Error while reading file: %s %s", path, err, exc_info=True
                    )
                    raise HTTPError(500, str(err))

            if not outputs.timestamp:
                set_kernelspec_from_language(model["content"])

            return model

        async def new_untitled(self, path="", type="", ext=""):
            """Create a new untitled file or directory in path

            We override the base function because that one does not take the 'ext' argument
            into account when type=="notebook". See https://github.com/mwouts/jupytext/issues/443
            """
            if type != "notebook" and ext != ".ipynb":
                return await self.super.new_untitled(path, type, ext)

            ext = ext or ".ipynb"
            if ":" in ext:
                ext, format_name = ext.split(":", 1)
            else:
                format_name = ""

            path = path.strip("/")
            if not await self.dir_exists(path):
                raise HTTPError(404, "No such directory: %s" % path)

            untitled = self.untitled_notebook
            config = await self.get_config(path)
            name = self.increment_notebook_filename(config, untitled + ext, path)
            path = f"{path}/{name}"

            model = {"type": "notebook"}
            if format_name:
                model["format"] = "json"
                model["content"] = nbformat.v4.nbbase.new_notebook(
                    metadata={"jupytext": {"formats": ext + ":" + format_name}}
                )

            return await self.new(model, path)

        def increment_notebook_filename(self, config, filename, path=""):
            """Increment a notebook filename until it is unique, regardless of extension"""
            # Extract the full suffix from the filename (e.g. .tar.gz)
            path = path.strip("/")
            basename, dot, ext = filename.partition(".")
            ext = dot + ext

            for i in itertools.count():
                if i:
                    insert_i = f"{i}"
                else:
                    insert_i = ""
                basename_i = basename + insert_i
                name = basename_i + ext
                if not any(
                    self.exists(f"{path}/{basename_i}{nb_ext}")
                    for nb_ext in config.notebook_extensions
                ):
                    break
            return name

        async def trust_notebook(self, path):
            """Trust the current notebook"""
            if path.endswith(".ipynb") or path not in self.paired_notebooks:
                await self.super.trust_notebook(path)
                return

            fmt, formats = self.paired_notebooks[path]
            for alt_path, alt_fmt in paired_paths(path, fmt, formats):
                if alt_fmt["extension"] == ".ipynb":
                    await self.super.trust_notebook(alt_path)

        async def rename_file(self, old_path, new_path):
            """Rename the current notebook, as well as its alternative representations"""
            if old_path not in self.paired_notebooks:
                try:
                    # we do not know yet if this is a paired notebook (#190)
                    # -> to get this information we open the notebook
                    await self.get(old_path, content=True)
                except Exception:
                    pass

            if old_path not in self.paired_notebooks:
                await self.super.rename_file(old_path, new_path)
                return

            fmt, formats = self.paired_notebooks.get(old_path)
            old_alt_paths = paired_paths(old_path, fmt, formats)

            # Is the new file name consistent with suffix?
            try:
                new_base = base_path(new_path, fmt)
            except HTTPError:
                raise
            except Exception as err:
                self.log.error(
                    "Error while renaming file from %s to %s: %s",
                    old_path,
                    new_path,
                    err,
                    exc_info=True,
                )
                raise HTTPError(500, str(err))

            for old_alt_path, alt_fmt in old_alt_paths:
                new_alt_path = full_path(new_base, alt_fmt)
                if self.exists(old_alt_path):
                    await self.create_prefix_dir(new_alt_path, alt_fmt)
                    await self.super.rename_file(old_alt_path, new_alt_path)

            self.drop_paired_notebook(old_path)
            self.update_paired_notebooks(new_path, formats)

        def get_parent_dir(self, path):
            """The parent directory"""
            if "/" in path:
                return path.rsplit("/", 1)[0]
            # jupyter-fs
            if ":" in path and hasattr(self, "_managers"):
                if path.endswith(":"):
                    return ""
                return path.rsplit(":", 1)[0] + ":"
            return ""

        async def get_config_file(self, directory):
            """Return the jupytext configuration file, if any"""
            for jupytext_config_file in JUPYTEXT_CONFIG_FILES:
                path = directory + "/" + jupytext_config_file
                if await self.file_exists(path):
                    if not self.allow_hidden and jupytext_config_file.startswith("."):
                        self.log.warning(
                            f"Ignoring config file {path} (see Jupytext issue #964)"
                        )
                        continue
                    return path

            pyproject_path = directory + "/" + PYPROJECT_FILE
            if await self.file_exists(pyproject_path):
                model = await self.get(pyproject_path, type="file")
                try:
                    doc = tomllib.loads(model["content"])
                except tomllib.TOMLDecodeError as e:
                    self.log.warning(f"Cannot load {pyproject_path}: {e}")
                else:
                    if doc.get("tool", {}).get("jupytext") is not None:
                        return pyproject_path

            if not directory:
                return None

            parent_dir = self.get_parent_dir(directory)
            return await self.get_config_file(parent_dir)

        async def load_config_file(
            self, config_file, *, prev_config_file, prev_config, is_os_path=False
        ):
            """Load the configuration file"""
            if config_file is None:
                return None
            if config_file.endswith(".py") and not is_os_path:
                config_file = self._get_os_path(config_file)
                is_os_path = True

            config_content = None
            if not is_os_path:
                try:
                    model = await self.super.get(config_file, content=True, type="file")
                    config_content = model["content"]
                except HTTPError:
                    pass

            config = load_jupytext_configuration_file(config_file, config_content)
            if config is None:
                return config

            log_level = config.cm_config_log_level
            if log_level == "info_if_changed":
                if config_file != prev_config_file or config != prev_config:
                    log_level = "info"
                else:
                    log_level = "none"
            if log_level != "none":
                getattr(self.log, log_level)(
                    "Loaded Jupytext configuration file at %s", config_file
                )
            return config

        async def get_config(self, path, use_cache=False):
            """Return the Jupytext configuration for the given path"""
            parent_dir = self.get_parent_dir(path)

            # When listing the notebooks for the tree view, we use a cache for the configuration file
            # The cache will be refreshed when a notebook is opened or saved, or when we go
            # to a different directory.
            if not use_cache or parent_dir != self.cached_config.path:
                try:
                    config_file = await self.get_config_file(parent_dir)
                    if config_file:
                        self.cached_config.config = await self.load_config_file(
                            config_file,
                            prev_config_file=self.cached_config.config_file,
                            prev_config=self.cached_config.config,
                        )
                    else:
                        config_file = find_global_jupytext_configuration_file()
                        self.cached_config.config = await self.load_config_file(
                            config_file,
                            prev_config_file=self.cached_config.config_file,
                            prev_config=self.cached_config.config,
                            is_os_path=True,
                        )
                    self.cached_config.config_file = config_file
                    self.cached_config.path = parent_dir
                except JupytextConfigurationError as err:
                    self.log.error(
                        "Error while reading config file: %s %s",
                        config_file,
                        err,
                        exc_info=True,
                    )
                    raise HTTPError(500, f"{err}")

            if self.cached_config.config is not None:
                return self.cached_config.config
            if isinstance(self.notebook_extensions, str):
                self.notebook_extensions = self.notebook_extensions.split(",")
            return self

    if "require_hash" in inspect.signature(base_contents_manager_class.get).parameters:
        AsyncJupytextContentsManager.get = (
            AsyncJupytextContentsManager._get_with_require_hash_argument
        )
    else:
        AsyncJupytextContentsManager.get = (
            AsyncJupytextContentsManager._get_with_no_require_hash_argument
        )

    return AsyncJupytextContentsManager


try:
    # The AsyncLargeFileManager is taken by default from jupyter_server if available
    from jupyter_server.services.contents.largefilemanager import AsyncLargeFileManager

    AsyncTextFileContentsManager = build_async_jupytext_contents_manager_class(
        AsyncLargeFileManager
    )
except ImportError:
    # If we can't find jupyter_server then we take it from notebook
    try:
        from notebook.services.contents.largefilemanager import AsyncLargeFileManager

        AsyncTextFileContentsManager = build_async_jupytext_contents_manager_class(
            AsyncLargeFileManager
        )
    except ImportError:
        # Older versions of notebook do not have the LargeFileManager #217
        from notebook.services.contents.filemanager import FileContentsManager

        AsyncTextFileContentsManager = build_async_jupytext_contents_manager_class(
            FileContentsManager
        )



import pytest
from nbformat.v4.nbbase import new_markdown_cell

from jupytext.cell_reader import (
    LightScriptCellReader,
    RMarkdownCellReader,
    paragraph_is_fully_commented,
    uncomment,
)
from jupytext.cell_to_text import RMarkdownCellExporter


@pytest.mark.parametrize(
    "lines",
    [
        "# text",
        """# # %%R
# # comment
# 1 + 1
# 2 + 2
""",
    ],
)
def test_paragraph_is_fully_commented(lines):
    assert paragraph_is_fully_commented(
        lines.splitlines(), comment="#", main_language="python"
    )


def test_paragraph_is_not_fully_commented(lines="# text\nnot fully commented out"):
    assert not paragraph_is_fully_commented(
        lines.splitlines(), comment="#", main_language="python"
    )


def test_uncomment():
    assert uncomment(["# line one", "#line two", "line three"], "#") == [
        "line one",
        "line two",
        "line three",
    ]
    assert uncomment(["# line one", "#line two", "line three"], "") == [
        "# line one",
        "#line two",
        "line three",
    ]


def test_text_to_code_cell():
    text = """```{python}
1+2+3
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "code"
    assert cell.source == "1+2+3"
    assert cell.metadata == {"language": "python"}
    assert lines[pos:] == []


def test_text_to_code_cell_empty_code():
    text = """```{python}
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "code"
    assert cell.source == ""
    assert cell.metadata == {"language": "python"}
    assert lines[pos:] == []


def test_text_to_code_cell_empty_code_no_blank_line():
    text = """```{python}
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "code"
    assert cell.source == ""
    assert cell.metadata == {"language": "python"}
    assert lines[pos:] == []


def test_text_to_markdown_cell():
    text = """This is
a markdown cell

```{python}
1+2+3
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "markdown"
    assert cell.source == "This is\na markdown cell"
    assert cell.metadata == {}
    assert pos == 3


def test_text_to_markdown_no_blank_line():
    text = """This is
a markdown cell
```{python}
1+2+3
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "markdown"
    assert cell.source == "This is\na markdown cell"
    assert cell.metadata == {"lines_to_next_cell": 0}
    assert pos == 2


def test_text_to_markdown_two_blank_line():
    text = """

```{python}
1+2+3
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "markdown"
    assert cell.source == ""
    assert cell.metadata == {}
    assert pos == 2


def test_text_to_markdown_one_blank_line():
    text = """
```{python}
1+2+3
```
"""
    lines = text.splitlines()
    cell, pos = RMarkdownCellReader().read(lines)

    assert cell.cell_type == "markdown"
    assert cell.source == ""
    assert cell.metadata == {"lines_to_next_cell": 0}
    assert pos == 1


def test_empty_markdown_to_text():
    cell = new_markdown_cell(source="")
    text = RMarkdownCellExporter(cell, "python").cell_to_text()
    assert text == [""]


def test_text_to_cell_py():
    text = "1+1\n"
    lines = text.splitlines()
    cell, pos = LightScriptCellReader().read(lines)
    assert cell.cell_type == "code"
    assert cell.source == "1+1"
    assert cell.metadata == {}
    assert pos == 1


def test_text_to_cell_py2():
    text = """def f(x):
    return x+1"""
    lines = text.splitlines()
    cell, pos = LightScriptCellReader().read(lines)
    assert cell.cell_type == "code"
    assert cell.source == """def f(x):\n    return x+1"""
    assert cell.metadata == {}
    assert pos == 2


def test_code_to_cell():
    text = """def f(x):
    return x+1"""
    lines = text.splitlines()
    cell, pos = LightScriptCellReader().read(lines)
    assert cell.cell_type == "code"
    assert cell.source == """def f(x):\n    return x+1"""
    assert cell.metadata == {}
    assert pos == 2


def test_uncomment_ocaml():
    assert uncomment(["(* ## *)"], "(*", "*)") == ["##"]
    assert uncomment(["(*##*)"], "(*", "*)") == ["##"]



from io import StringIO
from pathlib import Path

import nbformat
from nbformat.v4.nbbase import new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare, compare_notebooks


def test_simple_hook(tmpdir):
    nb_file = str(tmpdir.join("notebook.ipynb"))
    md_file = str(tmpdir.join("notebook.md"))
    nbformat.write(new_notebook(cells=[new_markdown_cell("Some text")]), nb_file)

    nb = jupytext.read(nb_file)
    jupytext.write(nb, md_file)

    with open(md_file) as fp:
        text = fp.read()

    assert "Some text" in text.splitlines()


def test_simple_hook_with_explicit_format(tmpdir):
    nb_file = str(tmpdir.join("notebook.ipynb"))
    py_file = str(tmpdir.join("notebook.py"))
    nbformat.write(new_notebook(cells=[new_markdown_cell("Some text")]), nb_file)

    nb = jupytext.read(nb_file)
    jupytext.write(nb, py_file, fmt="py:percent")

    with open(py_file) as fp:
        text = fp.read()

    assert "# %% [markdown]" in text.splitlines()
    assert "# Some text" in text.splitlines()


def test_no_error_on_path_object(tmpdir):
    nb_file = Path(str(tmpdir.join("notebook.ipynb")))
    md_file = nb_file.with_suffix(".md")
    nbformat.write(new_notebook(cells=[new_markdown_cell("Some text")]), str(nb_file))

    nb = jupytext.read(nb_file)
    jupytext.write(nb, md_file)


def test_read_ipynb_from_stream():
    def stream():
        return StringIO(
            """{
 "cells": [
  {
   "cell_type": "code",
   "metadata": {},
   "source": [
    "1 + 1"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 4
}
"""
        )

    nb = jupytext.read(stream())
    nb2 = jupytext.read(stream(), fmt="ipynb")
    compare(nb2, nb)


def test_read_py_percent_from_stream():
    def stream():
        return StringIO(
            """# %%
1 + 1
"""
        )

    nb = jupytext.read(stream())
    nb2 = jupytext.read(stream(), fmt="py:percent")
    compare_notebooks(nb2, nb)



import pytest

from jupytext.compare import assert_conversion_same_as_mirror

"""---------------------------------------------------------------------------------

Part I: ipynb -> fmt -> ipynb

---------------------------------------------------------------------------------"""


@pytest.mark.requires_pandoc
def test_ipynb_to_pandoc(ipynb_to_pandoc, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_to_pandoc, "md:pandoc", "ipynb_to_pandoc")


@pytest.mark.requires_quarto
def test_ipynb_to_quarto(
    ipynb_to_quarto,
    no_jupytext_version_number,
):
    assert_conversion_same_as_mirror(ipynb_to_quarto, "qmd", "ipynb_to_quarto")


@pytest.mark.requires_sphinx_gallery
def test_ipynb_to_python_sphinx(ipynb_to_sphinx, no_jupytext_version_number):
    assert_conversion_same_as_mirror(ipynb_to_sphinx, "py:sphinx", "ipynb_to_sphinx")


"""---------------------------------------------------------------------------------

Part II: text -> ipynb -> text

---------------------------------------------------------------------------------"""


def test_Rmd_to_ipynb(rmd_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(rmd_file, "ipynb", "Rmd_to_ipynb")


@pytest.mark.requires_sphinx_gallery
def test_sphinx_to_ipynb(sphinx_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(sphinx_file, "ipynb:sphinx", "sphinx_to_ipynb")


@pytest.mark.requires_sphinx_gallery
def test_sphinx_md_to_ipynb(sphinx_file, no_jupytext_version_number):
    assert_conversion_same_as_mirror(
        sphinx_file,
        {"extension": ".ipynb", "format_name": "sphinx", "rst2md": True},
        "sphinx-rst2md_to_ipynb",
        compare_notebook=True,
    )



from pathlib import Path

import pytest
from black import FileMode, format_str

from jupytext.async_contentsmanager import __file__ as async_contentsmanager_file
from jupytext.async_pairs import __file__ as async_pairs_file
from jupytext.sync_contentsmanager import __file__ as sync_contentsmanager_file
from jupytext.sync_pairs import __file__ as sync_pairs_file


def generate_async_code_from_sync_code(async_code: str) -> str:
    # No async methods
    sync_code = async_code.replace("async def", "def")

    # Don't await
    sync_code = sync_code.replace("await ", "")

    # Rename the contents manager
    sync_code = sync_code.replace("AsyncLargeFileManager", "LargeFileManager")
    sync_code = sync_code.replace(
        "AsyncJupytextContentsManager", "SyncJupytextContentsManager"
    )
    sync_code = sync_code.replace(
        "AsyncTextFileContentsManager", "TextFileContentsManager"
    )
    sync_code = sync_code.replace("asynchronous", "synchronous")
    sync_code = sync_code.replace("async_", "sync_")

    # Add a header
    sync_code = f'''"""
This file is automatically generated by
tests/functional/contents_manager/test_async_and_sync_contents_manager_are_in_sync.py
Do not edit this file manually.
"""

{sync_code}'''
    sync_code = format_str(sync_code, mode=FileMode())

    return sync_code


@pytest.mark.parametrize(
    "async_file, sync_file",
    [
        (async_contentsmanager_file, sync_contentsmanager_file),
        (async_pairs_file, sync_pairs_file),
    ],
)
def test_async_and_sync_files_are_in_sync(async_file, sync_file):
    async_code = Path(async_file).read_text()
    sync_code = generate_async_code_from_sync_code(async_code)

    actual_sync_code = Path(sync_file).read_text()
    try:
        assert actual_sync_code == sync_code
    except AssertionError:
        Path(sync_file).write_text(sync_code)
        raise



from copy import deepcopy

import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_markdown_cell
from pre_commit.main import main as pre_commit

import jupytext
from jupytext import TextFileContentsManager
from jupytext.compare import compare_notebooks


def test_pre_commit_hook_sync_with_no_config(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """In this test we reproduce the setting from https://github.com/mwouts/jupytext/issues/967"""
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [
      '--sync',
      '--set-formats',
      'ipynb,py:percent',
      '--show-changes',
      '--'
    ]
"""
    # Save a sample notebook with outputs in Jupyter
    nb = deepcopy(notebook_with_outputs)
    (tmpdir / "notebooks").mkdir()
    cm = TextFileContentsManager()
    cm.root_dir = str(tmpdir)
    cm.save(dict(type="notebook", content=nb), "nb.ipynb")

    # Add it to git
    tmp_repo.git.add("nb.ipynb")
    tmp_repo.index.commit("Notebook with outputs")

    # Configure the pre-commit hook
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)
    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    # Modify and save the notebook
    nb.cells.append(new_markdown_cell("New markdown cell"))
    cm.save(dict(type="notebook", content=nb), "nb.ipynb")

    # No .py file at the moment
    assert not (tmpdir / "nb.py").exists()

    # try to commit it, should fail as the py version hasn't been added
    tmp_repo.git.add("nb.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="git add nb.py",
    ):
        tmp_repo.index.commit("failing")

    # Now the .py file is present
    assert "New markdown cell" in tmpdir.join("nb.py").read()

    # add the ipynb and the py file, now the commit will succeed
    tmp_repo.git.add("nb.ipynb")
    tmp_repo.git.add("nb.py")
    tmp_repo.index.commit("passing")
    assert "nb.ipynb" in tmp_repo.tree()
    assert "nb.py" in tmp_repo.tree()

    # The notebook on disk is the same as the original, except for the new cell added
    nb = jupytext.read(tmpdir / "nb.ipynb")
    extra_cell = nb.cells.pop()
    assert extra_cell.cell_type == "markdown"
    assert extra_cell.source == "New markdown cell"
    compare_notebooks(
        nb,
        notebook_with_outputs,
        compare_ids=True,
        compare_outputs=True,
    )



from jupytext.cli import system


def test_jupytext_config_cli(tmp_path):
    settings_file = tmp_path / "default_setting_overrides.json"
    system("jupytext-config", "-h")
    system(
        "jupytext-config",
        "--settings-file",
        str(settings_file),
        "set-default-viewer",
        "python",
        "markdown",
    )
    assert "python" in (settings_file).read_text()
    assert "markdown" in system(
        "jupytext-config", "--settings-file", settings_file, "list-default-viewer"
    )
    system(
        "jupytext-config",
        "--settings-file",
        str(settings_file),
        "unset-default-viewer",
        "markdown",
    )
    list_viewers = system(
        "jupytext-config", "--settings-file", settings_file, "list-default-viewer"
    )
    assert "python" in list_viewers
    assert "markdown" not in list_viewers



from nbformat.v4.nbbase import new_markdown_cell, new_notebook, new_raw_cell

import jupytext
from jupytext.compare import compare
from jupytext.formats import get_format_implementation
from jupytext.header import (
    header_to_metadata_and_cell,
    metadata_and_cell_to_header,
    recursive_update,
    uncomment_line,
)


def test_uncomment():
    assert uncomment_line("# line one", "#") == "line one"
    assert uncomment_line("#line two", "#") == "line two"
    assert uncomment_line("#line two", "") == "#line two"


def test_header_to_metadata_and_cell_blank_line():
    text = """---
title: Sample header
---

Header is followed by a blank line
"""
    lines = text.splitlines()
    metadata, _, cell, pos = header_to_metadata_and_cell(lines, "", "")

    assert metadata == {}
    assert cell.cell_type == "raw"
    assert (
        cell.source
        == """---
title: Sample header
---"""
    )
    assert cell.metadata == {}
    assert lines[pos].startswith("Header is")


def test_header_to_metadata_and_cell_no_blank_line():
    text = """---
title: Sample header
---
Header is not followed by a blank line
"""
    lines = text.splitlines()
    metadata, _, cell, pos = header_to_metadata_and_cell(lines, "", "")

    assert metadata == {}
    assert cell.cell_type == "raw"
    assert (
        cell.source
        == """---
title: Sample header
---"""
    )
    assert cell.metadata == {"lines_to_next_cell": 0}
    assert lines[pos].startswith("Header is")


def test_header_to_metadata_and_cell_metadata():
    text = """---
title: Sample header
jupyter:
  mainlanguage: python
---
"""
    lines = text.splitlines()
    metadata, _, cell, pos = header_to_metadata_and_cell(lines, "", "")

    assert metadata == {"mainlanguage": "python"}
    assert cell.cell_type == "raw"
    assert (
        cell.source
        == """---
title: Sample header
---"""
    )
    assert cell.metadata == {"lines_to_next_cell": 0}
    assert pos == len(lines)


def test_metadata_and_cell_to_header(no_jupytext_version_number):
    metadata = {"jupytext": {"mainlanguage": "python"}}
    nb = new_notebook(
        metadata=metadata, cells=[new_raw_cell(source="---\ntitle: Sample header\n---")]
    )
    header, lines_to_next_cell = metadata_and_cell_to_header(
        nb, metadata, get_format_implementation(".md"), {"extension": ".md"}
    )
    assert (
        "\n".join(header)
        == """---
title: Sample header
jupyter:
  jupytext:
    mainlanguage: python
---"""
    )
    assert nb.cells == []
    assert lines_to_next_cell is None


def test_metadata_and_cell_to_header2(no_jupytext_version_number):
    nb = new_notebook(cells=[new_markdown_cell(source="Some markdown\ntext")])
    header, lines_to_next_cell = metadata_and_cell_to_header(
        nb, {}, get_format_implementation(".md"), {"extension": ".md"}
    )
    assert header == []
    assert len(nb.cells) == 1
    assert lines_to_next_cell is None


def test_notebook_from_plain_script_has_metadata_filter(
    script="""print('Hello world")
""",
):
    nb = jupytext.reads(script, ".py")
    assert nb.metadata.get("jupytext", {}).get("notebook_metadata_filter") == "-all"
    assert nb.metadata.get("jupytext", {}).get("cell_metadata_filter") == "-all"
    script2 = jupytext.writes(nb, ".py")

    compare(script2, script)


def test_multiline_metadata(
    no_jupytext_version_number,
    notebook=new_notebook(
        metadata={
            "multiline": """A multiline string

with a blank line""",
            "jupytext": {
                "notebook_metadata_filter": "all",
                "text_representation": {"extension": ".md", "format_name": "markdown"},
            },
        }
    ),
    markdown="""---
jupyter:
  jupytext:
    notebook_metadata_filter: all
    text_representation:
      extension: .md
      format_name: markdown
  multiline: 'A multiline string


    with a blank line'
---
""",
):
    actual = jupytext.writes(notebook, ".md")
    compare(actual, markdown)
    nb2 = jupytext.reads(markdown, ".md")
    compare(nb2, notebook)


def test_header_in_html_comment():
    text = """<!--

---
jupyter:
  title: Sample header
---

-->
"""
    lines = text.splitlines()
    metadata, _, cell, _ = header_to_metadata_and_cell(lines, "", "")

    assert metadata == {"title": "Sample header"}
    assert cell is None


def test_header_to_html_comment(no_jupytext_version_number):
    metadata = {"jupytext": {"mainlanguage": "python"}}
    nb = new_notebook(metadata=metadata, cells=[])
    header, lines_to_next_cell = metadata_and_cell_to_header(
        nb,
        metadata,
        get_format_implementation(".md"),
        {"extension": ".md", "hide_notebook_metadata": True},
    )
    compare(
        "\n".join(header),
        """<!--

---
jupyter:
  jupytext:
    mainlanguage: python
---

-->""",
    )


def test_recusive_update():
    assert recursive_update({0: {1: 2}}, {0: {1: 3}, 4: 5}) == {0: {1: 3}, 4: 5}
    assert recursive_update({0: {1: 2}}, {0: {1: 3}, 4: 5}, overwrite=False) == {
        0: {1: 2},
        4: 5,
    }
    # the value of `None`` is a special case
    assert recursive_update({0: 1}, {0: None}) == {}
    assert recursive_update({0: 1}, {0: None}, overwrite=False) == {}



import pytest

import jupytext
from jupytext.header import header_to_metadata_and_cell


@pytest.mark.requires_nbconvert
def test_markdown_jupytext_nbconvert_is_identity(md_file):
    """Test that a Markdown file, converted to a notebook, then
    exported back to Markdown with nbconvert, yields the original file"""
    if "jupytext" in md_file:
        pytest.skip()

    with open(md_file) as fp:
        md_org = fp.read()

    nb = jupytext.reads(md_org, "md")
    import nbconvert

    md_nbconvert, _ = nbconvert.export(nbconvert.MarkdownExporter, nb)

    # Our expectations
    md_expected = md_org.splitlines()
    # #region and #endregion comments are removed
    md_expected = [
        line
        for line in md_expected
        if line not in ["<!-- #region -->", "<!-- #endregion -->"]
    ]
    # language is not inserted by nbconvert
    md_expected = ["```" if line.startswith("```") else line for line in md_expected]
    # nbconvert inserts no empty line after the YAML header (which is in a Raw cell)
    md_expected = "\n".join(md_expected).replace("---\n\n", "---\n") + "\n"

    # an extra blank line is inserted before code cells
    md_nbconvert = md_nbconvert.replace("\n\n```", "\n```")

    jupytext.compare.compare(md_nbconvert, md_expected)


@pytest.mark.requires_nbconvert
def test_jupytext_markdown_similar_to_nbconvert(ipynb_py_R_jl_file):
    """Test that the nbconvert export for a notebook matches Jupytext's one"""
    if "magic" in ipynb_py_R_jl_file or "html" in ipynb_py_R_jl_file:
        pytest.skip()

    nb = jupytext.read(ipynb_py_R_jl_file)

    # Remove cell outputs and metadata
    for cell in nb.cells:
        if "outputs" in cell:
            cell.outputs = []
        if "metadata" in cell:
            cell.metadata = {}

    md_jupytext = jupytext.writes(nb, fmt="md")

    import nbconvert

    md_nbconvert, _ = nbconvert.export(nbconvert.MarkdownExporter, nb)

    # our expectations

    # nbconvert file has no YAML header
    md_jupytext_lines = md_jupytext.splitlines()
    _, _, raw_cell, pos = header_to_metadata_and_cell(md_jupytext_lines, "", "")
    md_jupytext = "\n".join(md_jupytext_lines[pos:]) + "\n"
    if raw_cell is not None:
        md_jupytext = raw_cell.source + "\n\n" + md_jupytext

    # region comments are not in nbconvert
    md_jupytext = md_jupytext.replace("<!-- #region -->\n", "").replace(
        "<!-- #endregion -->\n", ""
    )

    # Jupytext uses HTML comments to keep track of raw cells
    md_jupytext = (
        md_jupytext.replace("\n<!-- #raw -->\n", "")
        .replace("<!-- #raw -->\n", "")
        .replace("\n<!-- #endraw -->\n", "")
    )

    # nbconvert file may start with an empty line
    md_jupytext = md_jupytext.lstrip("\n")
    md_nbconvert = md_nbconvert.lstrip("\n")

    # Jupytext may not always have two blank lines between cells like Jupyter nbconvert
    md_jupytext = md_jupytext.replace("\n\n\n", "\n\n")
    md_nbconvert = md_nbconvert.replace("\n\n\n", "\n\n")

    jupytext.compare.compare(md_nbconvert, md_jupytext)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Part one - various cells

# Here we have a markdown cell
#
#
# with two blank lines

# Now we have a markdown cell
# with a code block inside it
#
# ```python
# 1 + 1
# ```
#
# After that cell we'll have a code cell

# +
2 + 2


3 + 3
# -

# Followed by a raw cell

# + active=""
# This is 
# the content
# of the raw cell
# -

# # Part two - cell metadata

# + [markdown] key="value"
# This is a markdown cell with cell metadata `{"key": "value"}`

# + .class tags=["parameters"]
"""This is a code cell with metadata `{"tags":["parameters"], ".class":null}`"""

# + key="value" active=""
# This is a raw cell with cell metadata `{"key": "value"}`



"""Read and write Jupyter notebooks as text files"""

import logging
import os
import sys
import warnings
from copy import copy, deepcopy

import nbformat
from nbformat.v4.nbbase import NotebookNode, new_code_cell, new_notebook
from nbformat.v4.rwbase import NotebookReader, NotebookWriter

from .cell_metadata import _IGNORE_CELL_METADATA
from .formats import (
    _VALID_FORMAT_OPTIONS,
    divine_format,
    format_name_for_ext,
    get_format_implementation,
    guess_format,
    long_form_one_format,
    read_format_from_metadata,
    rearrange_jupytext_metadata,
    update_jupytext_formats_metadata,
)
from .header import (
    _JUPYTER_METADATA_NAMESPACE,
    encoding_and_executable,
    header_to_metadata_and_cell,
    insert_jupytext_info_and_filter_metadata,
    insert_or_test_version_number,
    metadata_and_cell_to_header,
    metadata_and_cell_to_metadata,
    metadata_to_metadata_and_cell,
)
from .languages import (
    _SCRIPT_EXTENSIONS,
    default_language_from_metadata_and_ext,
    set_main_and_cell_language,
)
from .metadata_filter import filter_metadata, update_metadata_filters
from .myst import MYST_FORMAT_NAME, myst_extensions, myst_to_notebook, notebook_to_myst
from .pandoc import md_to_notebook, notebook_to_md
from .pep8 import pep8_lines_between_cells
from .quarto import notebook_to_qmd, qmd_to_notebook
from .version import __version__


class NotSupportedNBFormatVersion(NotImplementedError):
    """An error issued when the current notebook format is not supported by this version of Jupytext"""


class TextNotebookConverter(NotebookReader, NotebookWriter):
    """A class that can read or write a Jupyter notebook as text"""

    def __init__(self, fmt, config):
        self.fmt = copy(long_form_one_format(fmt))
        self.config = config
        self.ext = self.fmt["extension"]
        self.implementation = get_format_implementation(
            self.ext, self.fmt.get("format_name")
        )

    def update_fmt_with_notebook_options(self, metadata, read=False):
        """Update format options with the values in the notebook metadata, and record those
        options in the notebook metadata"""
        # The settings in the Jupytext configuration file have precedence over the metadata in the notebook
        # when the notebook is saved. This is because the metadata in the notebook might not be visible
        # in the text representation when e.g. notebook_metadata_filter="-all", which makes them hard to edit.
        if not read and self.config is not None:
            self.config.set_default_format_options(self.fmt, read)

        # Use format options from the notebook if not already set by the config
        for opt in _VALID_FORMAT_OPTIONS:
            if opt in metadata.get("jupytext", {}):
                self.fmt.setdefault(opt, metadata["jupytext"][opt])

        # When we read the notebook we use the values of the config as defaults, as again the text representation
        # of the notebook might not store the format options when notebook_metadata_filter="-all"
        if read and self.config is not None:
            self.config.set_default_format_options(self.fmt, read=read)

        # We save the format options in the notebook metadata
        for opt in _VALID_FORMAT_OPTIONS:
            if opt in self.fmt:
                metadata.setdefault("jupytext", {})[opt] = self.fmt[opt]

        # Is this format the same as that documented in the YAML header? If so, we want to know the format version
        file_fmt = metadata.get("jupytext", {}).get("text_representation", {})
        if self.fmt.get("extension") == file_fmt.get("extension") and self.fmt.get(
            "format_name"
        ) == file_fmt.get("format_name"):
            self.fmt.update(file_fmt)

        # rST to md conversion should happen only once
        if metadata.get("jupytext", {}).get("rst2md") is True:
            metadata["jupytext"]["rst2md"] = False

    def reads(self, s, **_):
        """Read a notebook represented as text"""
        if self.fmt.get("format_name") == "pandoc":
            return md_to_notebook(s)

        if self.fmt.get("format_name") == "quarto":
            return qmd_to_notebook(s)

        if self.fmt.get("format_name") == MYST_FORMAT_NAME:
            nb = myst_to_notebook(s)
            return self.split_frontmatter(nb)

        lines = s.splitlines()

        cells = []
        metadata, jupyter_md, header_cell, pos = header_to_metadata_and_cell(
            lines,
            self.implementation.header_prefix,
            self.implementation.header_suffix,
            self.implementation.extension,
            self.fmt.get(
                "root_level_metadata_as_raw_cell",
                self.config.root_level_metadata_as_raw_cell
                if self.config is not None
                else True,
            ),
        )
        default_language = default_language_from_metadata_and_ext(
            metadata, self.implementation.extension
        )
        self.update_fmt_with_notebook_options(metadata, read=True)

        if header_cell:
            cells.append(header_cell)

        lines = lines[pos:]

        if (
            self.implementation.format_name
            and self.implementation.format_name.startswith("sphinx")
        ):
            cells.append(new_code_cell(source="%matplotlib inline"))

        cell_metadata_json = False

        while lines:
            reader = self.implementation.cell_reader_class(self.fmt, default_language)
            cell, pos = reader.read(lines)
            cells.append(cell)
            cell_metadata_json = cell_metadata_json or reader.cell_metadata_json
            if pos <= 0:
                raise Exception(
                    "Blocked at lines " + "\n".join(lines[:6])
                )  # pragma: no cover
            lines = lines[pos:]

        custom_cell_magics = self.fmt.get("custom_cell_magics", "").split(",")
        set_main_and_cell_language(
            metadata, cells, self.implementation.extension, custom_cell_magics
        )
        cell_metadata = set()
        for cell in cells:
            cell_metadata.update(cell.metadata.keys())
        update_metadata_filters(metadata, jupyter_md, cell_metadata)

        if cell_metadata_json:
            metadata.setdefault("jupytext", {}).setdefault("cell_metadata_json", True)

        if (
            self.implementation.format_name
            and self.implementation.format_name.startswith("sphinx")
        ):
            filtered_cells = []
            for i, cell in enumerate(cells):
                if (
                    cell.source == ""
                    and i > 0
                    and i + 1 < len(cells)
                    and cells[i - 1].cell_type != "markdown"
                    and cells[i + 1].cell_type != "markdown"
                ):
                    continue
                filtered_cells.append(cell)
            cells = filtered_cells

        return new_notebook(cells=cells, metadata=metadata)

    def filter_notebook(self, nb, metadata, preserve_cell_ids=False):
        self.update_fmt_with_notebook_options(nb.metadata)
        unsupported_keys = set()
        metadata = insert_jupytext_info_and_filter_metadata(
            metadata, self.fmt, self.implementation, unsupported_keys=unsupported_keys
        )

        cells = []
        for cell in nb.cells:
            cell_metadata = filter_metadata(
                cell.metadata,
                self.fmt.get("cell_metadata_filter"),
                _IGNORE_CELL_METADATA,
                unsupported_keys=unsupported_keys,
            )

            if preserve_cell_ids and hasattr(cell, "id"):
                id = {"id": cell.id}
            else:
                id = {}

            if cell.cell_type == "code":
                cells.append(
                    new_code_cell(source=cell.source, metadata=cell_metadata, **id)
                )
            else:
                cells.append(
                    NotebookNode(
                        source=cell.source,
                        metadata=cell_metadata,
                        cell_type=cell.cell_type,
                        **id,
                    )
                )

        _warn_on_unsupported_keys(unsupported_keys)

        return NotebookNode(
            nbformat=nb.nbformat,
            nbformat_minor=nb.nbformat_minor,
            metadata=metadata,
            cells=cells,
        )

    def writes(self, nb, metadata=None, **kwargs):
        """Return the text representation of the notebook"""
        if self.fmt.get("format_name") == "pandoc":
            return notebook_to_md(
                self.filter_notebook(nb, metadata, preserve_cell_ids=True)
            )
        if self.fmt.get("format_name") == "quarto" or self.ext == ".qmd":
            return notebook_to_qmd(self.filter_notebook(nb, metadata))
        if self.fmt.get(
            "format_name"
        ) == MYST_FORMAT_NAME or self.ext in myst_extensions(no_md=True):
            default_lexer_from_language_info = metadata.get("language_info", {}).get(
                "pygments_lexer", None
            )
            default_lexer_from_jupytext_metadata = metadata.get("jupytext", {}).pop(
                "default_lexer", None
            )
            default_lexer = (
                default_lexer_from_language_info or default_lexer_from_jupytext_metadata
            )
            nb = self.filter_notebook(nb, metadata)
            nb = self.merge_frontmatter(nb)
            return notebook_to_myst(nb, default_lexer=default_lexer)

        # Copy the notebook, in order to be sure we do not modify the original notebook
        nb = NotebookNode(
            nbformat=nb.nbformat,
            nbformat_minor=nb.nbformat_minor,
            metadata=deepcopy(metadata or nb.metadata),
            cells=nb.cells,
        )

        metadata = nb.metadata
        default_language = (
            default_language_from_metadata_and_ext(
                metadata, self.implementation.extension, True
            )
            or "python"
        )
        self.update_fmt_with_notebook_options(nb.metadata)
        if "use_runtools" not in self.fmt:
            for cell in nb.cells:
                if cell.metadata.get("hide_input", False) or cell.metadata.get(
                    "hide_output", False
                ):
                    self.fmt["use_runtools"] = True
                    break

        header = encoding_and_executable(nb, metadata, self.ext)
        unsupported_keys = set()
        header_content, header_lines_to_next_cell = metadata_and_cell_to_header(
            nb,
            metadata,
            self.implementation,
            self.fmt,
            unsupported_keys=unsupported_keys,
        )
        header.extend(header_content)

        cell_exporters = []
        looking_for_first_markdown_cell = (
            self.implementation.format_name
            and self.implementation.format_name.startswith("sphinx")
        )
        split_at_heading = self.fmt.get("split_at_heading", False)

        for cell in nb.cells:
            if looking_for_first_markdown_cell and cell.cell_type == "markdown":
                cell.metadata.setdefault("cell_marker", '"""')
                looking_for_first_markdown_cell = False

            cell_exporters.append(
                self.implementation.cell_exporter_class(
                    cell, default_language, self.fmt, unsupported_keys=unsupported_keys
                )
            )

        _warn_on_unsupported_keys(unsupported_keys)

        texts = [cell.cell_to_text() for cell in cell_exporters]
        lines = []

        # concatenate cells in reverse order to determine how many blank lines (pep8)
        for i, cell in reversed(list(enumerate(cell_exporters))):
            text = cell.remove_eoc_marker(texts[i], lines)

            if (
                i == 0
                and self.implementation.format_name
                and self.implementation.format_name.startswith("sphinx")
                and (text in [["%matplotlib inline"], ["# %matplotlib inline"]])
            ):
                continue

            lines_to_next_cell = cell.lines_to_next_cell
            if lines_to_next_cell is None:
                lines_to_next_cell = pep8_lines_between_cells(
                    text, lines, self.implementation.extension
                )

            text.extend([""] * lines_to_next_cell)

            # two blank lines between markdown cells in Rmd when those do not have explicit region markers
            if self.ext in [".md", ".markdown", ".Rmd"] and not cell.is_code():
                if (
                    i + 1 < len(cell_exporters)
                    and not cell_exporters[i + 1].is_code()
                    and not texts[i][0].startswith("<!-- #")
                    and not texts[i + 1][0].startswith("<!-- #")
                    and (
                        not split_at_heading
                        or not (texts[i + 1] and texts[i + 1][0].startswith("#"))
                    )
                ):
                    text.append("")

            # "" between two consecutive code cells in sphinx
            if self.implementation.format_name.startswith("sphinx") and cell.is_code():
                if i + 1 < len(cell_exporters) and cell_exporters[i + 1].is_code():
                    text.append('""')

            lines = text + lines

        if header_lines_to_next_cell is None:
            header_lines_to_next_cell = pep8_lines_between_cells(
                header_content, lines, self.implementation.extension
            )

        header.extend([""] * header_lines_to_next_cell)

        return "\n".join(header + lines)

    def split_frontmatter(self, nb):
        """Use during self.reads to separate notebook metadata from other frontmatter."""
        unsupported_keys = set()
        metadata = nb.metadata.pop(_JUPYTER_METADATA_NAMESPACE, {})
        self.update_fmt_with_notebook_options(
            {
                "jupytext": {
                    **metadata.get("jupytext", {}),
                    **nb.metadata.get("jupytext", {}),
                },
            },
            read=True,
        )
        nb = metadata_to_metadata_and_cell(nb, metadata, self.fmt, unsupported_keys)
        _warn_on_unsupported_keys(unsupported_keys)
        return nb

    def merge_frontmatter(self, nb):
        """Use during self.writes to rewrite notebook metadata as frontmatter content."""
        unsupported_keys = set()
        nb = metadata_and_cell_to_metadata(nb, self.fmt, unsupported_keys)
        _warn_on_unsupported_keys(unsupported_keys)
        return nb


def reads(text, fmt=None, as_version=nbformat.NO_CONVERT, config=None, **kwargs):
    """
    Read a notebook from a string

    :param text: the text representation of the notebook
    :param fmt: (optional) the jupytext format like `md`, `py:percent`, ...
    :param as_version: see nbformat.reads
    :param config: (optional) a Jupytext configuration object
    :param kwargs: (not used) additional parameters for nbformat.reads
    :return: the notebook
    """
    fmt = copy(fmt) if fmt else divine_format(text)
    fmt = long_form_one_format(fmt)
    ext = fmt["extension"]

    if ext == ".ipynb":
        nb = nbformat.reads(text, as_version, **kwargs)
        (version, version_minor) = nbformat.reader.get_version(nb)
        if version != 4:
            warnings.warn(
                f"Notebooks in nbformat version {version}.{version_minor} are not supported by Jupytext. "
                f"Please consider converting them to nbformat version 4.x with "
                f"'jupyter nbconvert --to notebook --inplace'"
            )
        return nb

    format_name = read_format_from_metadata(text, ext) or fmt.get("format_name")

    if format_name:
        format_options = {}
    else:
        format_name, format_options = guess_format(text, ext)

    if format_name:
        fmt["format_name"] = format_name

    fmt.update(format_options)
    reader = TextNotebookConverter(fmt, config)
    notebook = reader.reads(text, **kwargs)
    rearrange_jupytext_metadata(notebook.metadata)

    if format_name and insert_or_test_version_number():
        notebook.metadata.setdefault("jupytext", {}).setdefault(
            "text_representation", {}
        ).update({"extension": ext, "format_name": format_name})

    return notebook


def read(fp, as_version=nbformat.NO_CONVERT, fmt=None, config=None, **kwargs):
    """Read a notebook from a file name or a file object

    :param fp: a file name or a file object
    :param as_version: see nbformat.read
    :param fmt: (optional) the jupytext format like `md`, `py:percent`, ...
    :param config: (optional) a Jupytext configuration object
    :param kwargs: (not used) additional parameters for nbformat.read
    :return: the notebook
    """
    if as_version != nbformat.NO_CONVERT and not isinstance(as_version, int):
        raise TypeError(
            "Second argument 'as_version' should be either nbformat.NO_CONVERT, or an integer."
        )

    if fp == "-":
        text = sys.stdin.read()
        # Update the input format by reference if missing
        if isinstance(fmt, dict) and not fmt:
            fmt.update(long_form_one_format(divine_format(text)))
        return reads(text, fmt)

    if not hasattr(fp, "read"):
        # Treat fp as a file name
        fp = str(fp)
        _, ext = os.path.splitext(fp)
        fmt = copy(fmt or {})
        if not isinstance(fmt, dict):
            fmt = long_form_one_format(fmt)
        fmt.update({"extension": ext})
        with open(fp, encoding="utf-8") as stream:
            return read(stream, as_version=as_version, fmt=fmt, config=config, **kwargs)

    if fmt is not None:
        fmt = long_form_one_format(fmt)
        if fmt["extension"] == ".ipynb":
            notebook = nbformat.read(fp, as_version, **kwargs)
            rearrange_jupytext_metadata(notebook.metadata)
            return notebook

    return reads(fp.read(), fmt, config=config, **kwargs)


def writes(notebook, fmt, version=nbformat.NO_CONVERT, config=None, **kwargs):
    """Return the text representation of the notebook

    :param notebook: the notebook
    :param fmt: the jupytext format like `md`, `py:percent`, ...
    :param version: see nbformat.writes
    :param config: (optional) a Jupytext configuration object
    :param kwargs: (not used) additional parameters for nbformat.writes
    :return: the text representation of the notebook
    """
    if version is not nbformat.NO_CONVERT:
        if not isinstance(version, int):
            raise TypeError(
                "The argument 'version' should be either nbformat.NO_CONVERT, or an integer."
            )
        notebook = nbformat.convert(notebook, version)
    (version, version_minor) = nbformat.reader.get_version(notebook)
    if version < 4:
        raise NotSupportedNBFormatVersion(
            f"Notebooks in nbformat version {version}.{version_minor} are not supported by Jupytext. "
            f"Please convert your notebooks to nbformat version 4 with "
            f"'jupyter nbconvert --to notebook --inplace', or call this function with 'version=4'."
        )
    if version > 4 or (version == 4 and version_minor > 5):
        warnings.warn(
            f"Notebooks in nbformat version {version}.{version_minor} "
            f"have not been tested with Jupytext version {__version__}."
        )

    metadata = deepcopy(notebook.metadata)
    rearrange_jupytext_metadata(metadata)
    fmt = copy(fmt)
    fmt = long_form_one_format(fmt, metadata)
    ext = fmt["extension"]
    format_name = fmt.get("format_name")

    if ext == ".ipynb":
        return nbformat.writes(
            drop_text_representation_metadata(notebook, metadata),
            version,
            **kwargs,
        )

    if not format_name:
        format_name = format_name_for_ext(metadata, ext, explicit_default=False)

    # Since Jupytext==1.17, the default format for
    # writing a notebook to a script is the percent format
    if not format_name and "cell_markers" not in fmt and ext in _SCRIPT_EXTENSIONS:
        format_name = "percent"

    if format_name:
        fmt["format_name"] = format_name
        update_jupytext_formats_metadata(metadata, fmt)

    writer = TextNotebookConverter(fmt, config)
    return writer.writes(notebook, metadata)


def drop_text_representation_metadata(notebook, metadata=None):
    """When the notebook is saved to an ipynb file, we drop the text_representation metadata"""
    if metadata is None:
        # Make a copy to avoid modification by reference
        metadata = deepcopy(notebook["metadata"])

    jupytext_metadata = metadata.get("jupytext", {})
    jupytext_metadata.pop("text_representation", {})

    # Remove the jupytext section if empty
    if not jupytext_metadata:
        metadata.pop("jupytext", {})

    return NotebookNode(
        nbformat=notebook["nbformat"],
        nbformat_minor=notebook["nbformat_minor"],
        metadata=metadata,
        cells=notebook["cells"],
    )


def write(nb, fp, version=nbformat.NO_CONVERT, fmt=None, config=None, **kwargs):
    """Write a notebook to a file name or a file object

    :param nb: the notebook
    :param fp: a file name or a file object
    :param version: see nbformat.write
    :param fmt: (optional if fp is a file name) the jupytext format like `md`, `py:percent`, ...
    :param config: (optional) a Jupytext configuration object
    :param kwargs: (not used) additional parameters for nbformat.write
    """
    if fp == "-":
        # Use sys.stdout.buffer when possible, and explicit utf-8 encoding, cf. #331
        content = writes(nb, version=version, fmt=fmt, config=config, **kwargs)
        try:
            # Python 3
            sys.stdout.buffer.write(content.encode("utf-8"))
        except AttributeError:
            sys.stdout.write(content.encode("utf-8"))
        return

    if not hasattr(fp, "write"):
        # Treat fp as a file name
        fp = str(fp)
        _, ext = os.path.splitext(fp)
        fmt = copy(fmt or {})
        fmt = long_form_one_format(fmt, update={"extension": ext})
        create_prefix_dir(fp, fmt)

        with open(fp, "w", encoding="utf-8") as stream:
            write(nb, stream, version=version, fmt=fmt, config=config, **kwargs)
            return
    else:
        assert (
            fmt is not None
        ), "'fmt' argument in jupytext.write is mandatory unless fp is a file name"

    content = writes(nb, version=version, fmt=fmt, config=config, **kwargs)
    if isinstance(content, bytes):
        content = content.decode("utf8")
    fp.write(content)
    if not content.endswith("\n"):
        fp.write("\n")


def create_prefix_dir(nb_file, fmt):
    """Create directory if fmt has a prefix"""
    if "prefix" in fmt:
        nb_dir = os.path.dirname(nb_file) + os.path.sep
        if not os.path.isdir(nb_dir):
            logging.log(
                logging.WARNING, "[jupytext] creating missing directory %s", nb_dir
            )
            os.makedirs(nb_dir)


def _warn_on_unsupported_keys(unsupported_keys):
    if unsupported_keys:
        warnings.warn(
            f"The following metadata cannot be exported "
            f"to the text notebook: {sorted(unsupported_keys)}"
        )



# ---
# title: Quick test
# output:
#   ioslides_presentation:
#     widescreen: true
#     smaller: true
# editor_options:
#   chunk_output_type: console
# jupyter:
#   jupytext:
#     cell_markers: '{{{,}}}'
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

1+2+3





"""Functions to read or write paired notebooks"""

from collections import namedtuple

from .formats import long_form_multiple_formats, long_form_one_format
from .paired_paths import paired_paths

NotebookFile = namedtuple("notebook_file", "path fmt timestamp")


class PairedFilesDiffer(ValueError):
    """An error when the two representations of a paired notebook differ"""


def latest_inputs_and_outputs(
    path, fmt, formats, get_timestamp, contents_manager_mode=False
):
    """Given a notebook path, its format and paired formats, and a function that
    returns the timestamp for each (or None if the file does not exist), return
    the most recent notebook for the inputs and outputs, respectively"""

    timestamp_inputs = None
    timestamp_outputs = None
    inputs_path = None
    outputs_path = None
    input_fmt = None
    output_fmt = None

    fmt = long_form_one_format(fmt)
    formats = long_form_multiple_formats(formats)

    for alt_path, alt_fmt in paired_paths(path, fmt, formats):
        # In the contents manager, we don't read another text file if the current notebook is in already in text mode
        if (
            contents_manager_mode
            and alt_fmt["extension"] != ".ipynb"
            and fmt["extension"] != ".ipynb"
        ):
            if any(
                alt_fmt.get(key) != fmt.get(key)
                for key in ["extension", "suffix", "prefix"]
            ):
                continue

        timestamp = get_timestamp(alt_path)
        if timestamp is None:
            continue

        if alt_fmt["extension"] == ".ipynb":
            if timestamp_outputs is None or timestamp > timestamp_outputs:
                timestamp_outputs = timestamp
                outputs_path, output_fmt = alt_path, alt_fmt
        elif timestamp_inputs is None or timestamp > timestamp_inputs:
            timestamp_inputs = timestamp
            inputs_path, input_fmt = alt_path, alt_fmt

    if timestamp_inputs is None or (
        not contents_manager_mode
        and timestamp_outputs
        and timestamp_outputs > timestamp_inputs
    ):
        timestamp_inputs = timestamp_outputs
        inputs_path, input_fmt = outputs_path, output_fmt

    return (
        NotebookFile(inputs_path, input_fmt, timestamp_inputs),
        NotebookFile(outputs_path, output_fmt, timestamp_outputs),
    )



# In this notebook, we benchmark the Jupytext formats for Jupyter notebooks against the base format
# Open this script as a notebook in Jupyter to run it and see the plots

import time
import copy
import pandas as pd
import plotly.graph_objects as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

import nbformat
import jupytext

# The notebook to be tested
notebook = jupytext.read('World population.ipynb')

# Same notebook, with no outputs, for a fair comparison
notebook_no_outputs = copy.deepcopy(notebook)
for cell in notebook_no_outputs.cells:
    cell.outputs = []
    cell.execution_count = None

# +
JUPYTEXT_FORMATS = ['ipynb', 'md', 'py:light', 'py:percent', 'py:sphinx']

# Let's see if we have pandoc here
try:
    jupytext.writes(notebook, fmt='md:pandoc')
    JUPYTEXT_FORMATS.append('md:pandoc')
except jupytext.formats.JupytextFormatError as err:
    print(str(err))


# Let's see if we have myst-parser installed here
try:
    jupytext.writes(notebook, fmt='myst')
    JUPYTEXT_FORMATS.append('myst')
except jupytext.formats.JupytextFormatError as err:
    print(str(err))

# -


def sample_perf(nb, n=30):
    samples = pd.DataFrame(
        pd.np.NaN,
        index=pd.MultiIndex.from_product(
            (range(n), ['nbformat'] + JUPYTEXT_FORMATS), names=['sample', 'implementation']),
        columns=pd.Index(['size', 'read', 'write'], name='measure'))

    for i, fmt in samples.index:
        t0 = time.time()
        if fmt == 'nbformat':
            text = nbformat.writes(nb)
        else:
            text = jupytext.writes(nb, fmt)
        t1 = time.time()
        samples.loc[(i, fmt), 'write'] = t1 - t0
        samples.loc[(i, fmt), 'size'] = len(text)
        t0 = time.time()
        if fmt == 'nbformat':
            nbformat.reads(text, as_version=4)
        else:
            jupytext.reads(text, fmt)
        t1 = time.time()
        samples.loc[(i, fmt), 'read'] = t1 - t0
    return samples


def performance_plot(perf, title):
    formats = ['nbformat'] + JUPYTEXT_FORMATS
    mean = perf.groupby('implementation').mean().loc[formats]
    std = perf.groupby('implementation').std().loc[formats]
    data = [go.Bar(x=mean.index,
                   y=mean[col],
                   error_y=dict(
                       type='data',
                       array=std[col],
                       color=color,
                       thickness=0.5
                   ) if col != 'size' else dict(),
                   name=col,
                   yaxis={'read': 'y1', 'write': 'y2', 'size': 'y3'}[col])
            for col, color in zip(mean.columns, DEFAULT_PLOTLY_COLORS)]
    layout = go.Layout(title=title,
                       xaxis=dict(title='Implementation', anchor='y3'),
                       yaxis=dict(domain=[0.7, 1], title='Read (secs)'),
                       yaxis2=dict(domain=[0.35, .65], title='Write (secs)'),
                       yaxis3=dict(domain=[0, .3], title='Size')
                       )
    return go.Figure(data=data, layout=layout)


perf_no_outputs = sample_perf(notebook_no_outputs, 30)

performance_plot(perf_no_outputs, 'Benchmarking Jupytext on the World Population notebook<br>(Outputs filtered)')

perf = sample_perf(notebook, 30)

performance_plot(perf, 'Benchmarking Jupytext on the World Population notebook<br>(With outputs)')



from jupytext.stringparser import StringParser


def test_long_string(
    text="""'''This is a multiline
comment with "quotes", 'single quotes'
# and comments
and line breaks


and it ends here'''


1 + 1
""",
):
    quoted = []
    sp = StringParser("python")
    for i, line in enumerate(text.splitlines()):
        if sp.is_quoted():
            quoted.append(i)
        sp.read_line(line)

    assert quoted == [1, 2, 3, 4, 5, 6]


def test_single_chars(
    text="""'This is a single line comment'''
'and another one'
# and comments
"and line breaks"


"and it ends here'''"


1 + 1
""",
):
    sp = StringParser("python")
    for line in text.splitlines():
        assert not sp.is_quoted()
        sp.read_line(line)


def test_long_string_with_four_quotes(
    text="""''''This is a multiline
comment that starts with four quotes
'''

1 + 1
""",
):
    quoted = []
    sp = StringParser("python")
    for i, line in enumerate(text.splitlines()):
        if sp.is_quoted():
            quoted.append(i)
        sp.read_line(line)

    assert quoted == [1, 2]


def test_long_string_ends_with_four_quotes(
    text="""'''This is a multiline
comment that ends with four quotes
''''

1 + 1
""",
):
    quoted = []
    sp = StringParser("python")
    for i, line in enumerate(text.splitlines()):
        if sp.is_quoted():
            quoted.append(i)
        sp.read_line(line)

    assert quoted == [1, 2]



import os
import shutil

import pytest
from jupyter_server.utils import ensure_async
from nbformat.v4.nbbase import new_code_cell, new_output

from jupytext.compare import compare_notebooks

pytestmark = pytest.mark.asyncio


async def test_py_notebooks_are_trusted(python_file, cm):
    root, file = os.path.split(python_file)
    cm.root_dir = root
    nb = await ensure_async(cm.get(file))
    for cell in nb["content"].cells:
        assert cell.metadata.get("trusted", True)


async def test_rmd_notebooks_are_trusted(rmd_file, cm):
    root, file = os.path.split(rmd_file)
    cm.root_dir = root
    nb = await ensure_async(cm.get(file))
    for cell in nb["content"].cells:
        assert cell.metadata.get("trusted", True)


@pytest.mark.skip(reason="Fails intermittently on CI, see #1346")
async def test_ipynb_notebooks_can_be_trusted(
    ipynb_py_file, tmpdir, no_jupytext_version_number, cm
):
    if "hash sign" in ipynb_py_file:
        pytest.skip()
    root, file = os.path.split(ipynb_py_file)
    tmp_ipynb = str(tmpdir.join(file))
    py_file = file.replace(".ipynb", ".py")
    tmp_py = str(tmpdir.join(py_file))
    shutil.copy(ipynb_py_file, tmp_ipynb)

    cm.formats = "ipynb,py"
    cm.root_dir = str(tmpdir)
    model = await ensure_async(cm.get(file))
    await ensure_async(cm.save(model, py_file))

    # Unsign and test notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    model = await ensure_async(cm.get(file))
    for cell in model["content"].cells:
        assert (
            "trusted" not in cell.metadata
            or not cell.metadata["trusted"]
            or not cell.outputs
        )

    # Trust and reload
    await ensure_async(cm.trust_notebook(py_file))

    model = await ensure_async(cm.get(file))
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)

    # Remove py file, content should be the same
    os.remove(tmp_py)
    nb2 = await ensure_async(cm.get(file))
    for cell in nb2["content"].cells:
        assert cell.metadata.get("trusted", True)

    compare_notebooks(nb2["content"], model["content"])

    # Just for coverage
    await ensure_async(cm.trust_notebook(file))


@pytest.mark.skip(reason="Fails intermittently on CI, see #1346")
async def test_ipynb_notebooks_can_be_trusted_even_with_metadata_filter(
    ipynb_py_file, tmpdir, no_jupytext_version_number, cm
):
    if "hash sign" in ipynb_py_file:
        pytest.skip()
    root, file = os.path.split(ipynb_py_file)
    tmp_ipynb = str(tmpdir.join(file))
    py_file = file.replace(".ipynb", ".py")
    tmp_py = str(tmpdir.join(py_file))
    shutil.copy(ipynb_py_file, tmp_ipynb)

    cm.formats = "ipynb,py"
    cm.notebook_metadata_filter = "all"
    cm.cell_metadata_filter = "-all"
    cm.root_dir = str(tmpdir)
    model = await ensure_async(cm.get(file))
    await ensure_async(cm.save(model, py_file))

    # Unsign notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    # Trust and reload
    await ensure_async(cm.trust_notebook(py_file))

    model = await ensure_async(cm.get(file))
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)

    # Remove py file, content should be the same
    os.remove(tmp_py)
    nb2 = await ensure_async(cm.get(file))
    for cell in nb2["content"].cells:
        assert cell.metadata.get("trusted", True)

    compare_notebooks(nb2["content"], model["content"])


async def test_text_notebooks_can_be_trusted(
    percent_file, tmpdir, no_jupytext_version_number, cm
):
    root, file = os.path.split(percent_file)
    py_file = str(tmpdir.join(file))
    shutil.copy(percent_file, py_file)

    cm.root_dir = str(tmpdir)
    model = await ensure_async(cm.get(file))
    model["type"] == "notebook"
    await ensure_async(cm.save(model, file))

    # Unsign notebook
    nb = model["content"]
    for cell in nb.cells:
        cell.metadata.pop("trusted", True)

    cm.notary.unsign(nb)

    # Trust and reload
    await ensure_async(cm.trust_notebook(file))

    model = await ensure_async(cm.get(file))
    for cell in model["content"].cells:
        assert cell.metadata.get("trusted", True)


async def test_simple_notebook_is_trusted(tmpdir, python_notebook, cm):
    cm.root_dir = str(tmpdir)

    nb = python_notebook
    cm.notary.unsign(nb)

    # All cells are trusted in this notebook
    assert cm.notary.check_cells(nb)
    # Yet the notebook is not in the database of trusted notebooks
    assert not cm.notary.check_signature(nb)

    # Save the notebook using the CM
    await ensure_async(cm.save(dict(type="notebook", content=nb), "test.ipynb"))

    # The notebook is safe so it should have been trusted before getting saved to disk
    nb = (await ensure_async(cm.get("test.ipynb")))["content"]
    assert cm.notary.check_signature(nb)


@pytest.mark.requires_myst
async def test_myst_notebook_is_trusted_941(
    tmp_path,
    cm,
    myst="""---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.11.5
kernelspec:
  display_name: itables
  language: python
  name: itables
---

# Downsampling

```{code-cell} ipython3
from itables import init_notebook_mode, show

init_notebook_mode(all_interactive=True)
```
""",
):
    cm.root_dir = str(tmp_path)

    test_md = tmp_path / "test.md"
    test_md.write_text(myst)

    nb = (await ensure_async(cm.get("test.md")))["content"]

    # All cells are trusted in this notebook
    assert cm.notary.check_cells(nb)


@pytest.mark.requires_myst
async def test_paired_notebook_with_outputs_is_not_trusted_941(
    tmp_path, python_notebook, cm
):
    cm.root_dir = str(tmp_path)

    nb = python_notebook
    nb.cells.append(new_code_cell(source="1+1", outputs=[new_output("execute_result")]))
    nb.metadata["jupytext"] = {"formats": "ipynb,md:myst"}
    cm.notary.unsign(nb)
    await ensure_async(
        cm.save(model=dict(type="notebook", content=nb), path="test.ipynb")
    )

    nb = (await ensure_async(cm.get("test.md")))["content"]
    assert not cm.notary.check_cells(nb)
    assert not cm.notary.check_signature(nb)



import re
from time import sleep

import nbformat
import pytest
from nbformat.v4.nbbase import (
    new_code_cell,
    new_markdown_cell,
    new_notebook,
    new_raw_cell,
)

import jupytext
from jupytext.cli import jupytext as jupytext_cli
from jupytext.compare import compare, compare_cells, compare_notebooks


def test_read_mostly_py_rmd_file(
    rmd="""---
title: Simple file
---

```{python, echo=TRUE}
import numpy as np
x = np.arange(0, 2*math.pi, eps)
```

```{python, echo=TRUE}
x = np.arange(0,1,eps)
y = np.abs(x)-.5
```

```{r}
ls()
```

```{r, results="asis", magic_args="-i x"}
cat(stringi::stri_rand_lipsum(3), sep='\n\n')
```
""",
):
    nb = jupytext.reads(rmd, "Rmd")
    compare_cells(
        nb.cells,
        [
            new_raw_cell("---\ntitle: Simple file\n---"),
            new_code_cell(
                "import numpy as np\n" "x = np.arange(0, 2*math.pi, eps)",
                metadata={"echo": True},
            ),
            new_code_cell(
                "x = np.arange(0,1,eps)\ny = np.abs(x)-.5", metadata={"echo": True}
            ),
            new_code_cell("%%R\nls()"),
            new_code_cell(
                "%%R -i x\ncat(stringi::" "stri_rand_lipsum(3), sep='\n\n')",
                metadata={"results": "asis"},
            ),
        ],
        compare_ids=False,
    )

    rmd2 = jupytext.writes(nb, "Rmd")
    rmd2 = re.sub(r"```{r ", "```{r, ", rmd2)
    rmd2 = re.sub(r"```{python ", "```{python, ", rmd2)
    compare(rmd2, rmd)


def test_markdown_cell_with_code_works(
    nb=new_notebook(
        cells=[
            new_markdown_cell(
                """```python
1 + 1
```"""
            )
        ]
    ),
):
    text = jupytext.writes(nb, "Rmd")
    nb2 = jupytext.reads(text, "Rmd")
    compare_notebooks(nb2, nb)


def test_two_markdown_cell_with_code_works(
    nb=new_notebook(
        cells=[
            new_markdown_cell(
                """```python
1 + 1
```"""
            ),
            new_markdown_cell(
                """```python
2 + 2
```"""
            ),
        ]
    )
):
    text = jupytext.writes(nb, "Rmd")
    nb2 = jupytext.reads(text, "Rmd")
    compare_notebooks(nb2, nb)


def test_tags_in_rmd(
    rmd="""---
jupyter:
  jupytext:
    text_representation:
      extension: .Rmd
      format_name: rmarkdown
      format_version: '1.1'
      jupytext_version: 1.2.3
---

```{python tags=c("parameters")}
p = 1
```
""",
    nb=new_notebook(cells=[new_code_cell("p = 1", metadata={"tags": ["parameters"]})]),
):
    nb2 = jupytext.reads(rmd, "Rmd")
    compare_notebooks(nb2, nb)


def round_trip_cell_metadata(cell_metadata):
    nb = new_notebook(
        metadata={"jupytext": {"main_language": "python"}},
        cells=[new_code_cell("1 + 1", metadata=cell_metadata)],
    )
    text = jupytext.writes(nb, "Rmd")
    nb2 = jupytext.reads(text, "Rmd")
    compare_notebooks(nb2, nb)


def test_comma_in_metadata(cell_metadata={"a": "b, c"}):
    round_trip_cell_metadata(cell_metadata)


def test_dict_in_metadata(cell_metadata={"a": {"b": "c"}}):
    round_trip_cell_metadata(cell_metadata)


def test_list_in_metadata(cell_metadata={"d": ["e"]}):
    round_trip_cell_metadata(cell_metadata)


@pytest.mark.parametrize("root_level_metadata_as_raw_cell", [True, False])
def test_root_level_metadata_as_raw_cell(
    tmpdir,
    root_level_metadata_as_raw_cell,
    rmd="""---
author: R Markdown document author
title: R Markdown notebook title
---

```{r}
1 + 1
```
""",
):
    nb_file = tmpdir.join("notebook.ipynb")
    rmd_file = tmpdir.join("notebook.Rmd")
    cfg_file = tmpdir.join("jupytext.toml")

    cfg_file.write(
        "root_level_metadata_as_raw_cell = {}".format(
            "true" if root_level_metadata_as_raw_cell else "false"
        )
    )
    rmd_file.write(rmd)

    jupytext_cli([str(rmd_file), "--to", "ipynb"])

    nb = nbformat.read(str(nb_file), as_version=4)

    if root_level_metadata_as_raw_cell:
        compare_cells(
            nb.cells,
            [
                new_raw_cell(
                    """---
author: R Markdown document author
title: R Markdown notebook title
---"""
                ),
                new_code_cell("1 + 1"),
            ],
            compare_ids=False,
        )
    else:
        compare_cells(nb.cells, [new_code_cell("1 + 1")], compare_ids=False)
        assert nb.metadata["jupytext"]["root_level_metadata"] == {
            "title": "R Markdown notebook title",
            "author": "R Markdown document author",
        }

    # Writing back to Rmd should preserve the original document
    jupytext_cli([str(nb_file), "--to", "Rmd"])
    compare(rmd_file.read(), rmd)


def test_pair_rmd_file_with_cell_tags_and_options(
    tmpdir, cwd_tmpdir, no_jupytext_version_number
):
    rmd = """```{r plot_1, dpi=72}
plot(3:30)
```
"""
    rmd_file = tmpdir.join("test.Rmd")
    rmd_file.write(rmd)

    # Pair Rmd with ipynb
    jupytext_cli(["--set-formats", "ipynb,Rmd", "test.Rmd"])

    # Wait so that ipynb will be more recent
    sleep(0.2)

    # Modify the ipynb file
    nb_file = tmpdir.join("test.ipynb")
    nb = nbformat.read(nb_file, as_version=4)
    nb.cells[0].source = "plot(4:40)"
    nb_file.write(nbformat.writes(nb))

    # Sync the two files
    jupytext_cli(["--sync", "test.Rmd"])

    # Remove the header
    rmd2 = rmd_file.read()
    rmd2 = rmd2.rsplit("---\n\n")[1]

    # The new code chunk has the new code, and options are still there
    compare(rmd2, rmd.replace("3", "4"))


def test_apostrophe_in_parameter_1079(
    rmd="""```{python some-name, param="Problem's"}
a = 1
```
""",
):
    nb = jupytext.reads(rmd, fmt="Rmd")
    rmd2 = jupytext.writes(nb, fmt="Rmd")
    compare(rmd2, rmd)
    nb2 = jupytext.reads(rmd, fmt="Rmd")
    compare_notebooks(nb2, nb)


@pytest.mark.parametrize("line", [" #'''", "a = 2  #'''"])
def test_commented_triple_quote_1060(line):
    rmd = f"""```{{python}}
{line}
```

```{{python}}
# Another cell
```
"""
    nb = jupytext.reads(rmd, fmt="Rmd")
    assert nb.cells[0].source == line

    rmd2 = jupytext.writes(nb, fmt="Rmd")
    compare(rmd2, rmd)
    nb2 = jupytext.reads(rmd, fmt="Rmd")
    compare_notebooks(nb2, nb)


def test_bibliography_in_rmd(
    rmd="""Issue #1161

The bibliography section below should not
become a code cell

```{bibliography}
```

```{r}
6
```
""",
):
    nb = jupytext.reads(rmd, fmt="Rmd")
    assert len(nb.cells) == 2, nb.cells
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "6"
    rmd2 = jupytext.writes(nb, fmt="Rmd")
    compare(rmd2, rmd)



# ---
# jupyter:
#   jupytext:
#     cell_markers: region,endregion
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

1+2+3

# region active=""
# This is a raw cell
# endregion

# This is a markdown cell



# ---
# jupyter:
#   jupytext:
#     cell_markers: '{{{,}}}'
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# {{{ raw_mimetype="text/latex" active=""
# $1+1$
# }}}

# {{{ raw_mimetype="text/restructuredtext" active=""
# :math:`1+1`
# }}}

# {{{ raw_mimetype="text/html" active=""
# <b>Bold text<b>
# }}}

# {{{ raw_mimetype="text/markdown" active=""
# **Bold text**
# }}}

# {{{ raw_mimetype="text/x-python" active=""
# 1 + 1
# }}}

# {{{ active=""
# Not formatted
# }}}



import pytest
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell

from jupytext import reads, writes
from jupytext.formats import formats_with_support_for_cell_metadata, is_myst_available


@pytest.fixture()
def notebook_with_tags(python_notebook):
    nb = python_notebook
    nb.cells = [
        new_code_cell("1 + 1", metadata={"tags": ["tag1"]}),
        new_markdown_cell("some text", metadata={"tags": ["tag2"]}),
    ]
    return nb


@pytest.mark.parametrize(
    "fmt", ["py:percent", "py:light", "md:markdown", "md:myst", "Rmd:rmarkdown"]
)
def test_main_formats_support_cell_metadata(fmt):
    if fmt == "md:myst" and not is_myst_available():
        pytest.skip("myst is not available")
    assert fmt in set(formats_with_support_for_cell_metadata())


def test_tags_are_preserved(notebook_with_tags, fmt_with_cell_metadata):
    text = writes(notebook_with_tags, fmt_with_cell_metadata)
    nb = reads(text, fmt_with_cell_metadata)
    assert nb.cells[0]["metadata"]["tags"] == ["tag1"]
    assert nb.cells[1]["metadata"]["tags"] == ["tag2"]



# This is a notebook that I used to generate Jupytext's word cloud.
# To open this script as a notebook in JupyterLab, right-click on this file, and select _Open with/Notebook_.

from wordcloud import WordCloud

text = """
Jupytext
Notebook
JupyterLab
Git
GitHub
Version control
Markdown
R Markdown
Text
Scripts
Code
Notebook Template
Binder
Visual Studio Code
PyCharm
Atom
Spyder
Hydrogen
RStudio
Sphinx-Gallery
Documentation
black
pytest
autopep8
Metadata
Reproducible research
R
Julia
Python
Bash
Powershell
Scala
Scheme
Clojure
Matlab
Octave
C++
q/kdb+
IDL
TypeScript
Javascript
Scala
Rust
Robot Framework
"""

wordcloud = WordCloud(
    random_state=1,
    background_color='white',
    width=1200, height=500
).generate_from_frequencies({word: 1 for word in text.splitlines()})

wordcloud.to_image()

wordcloud.to_file('../docs/jupytext_word_cloud.png')



import re

import pytest
from jupyter_server.utils import ensure_async

import jupytext
from jupytext.compare import compare_notebooks, notebook_model

pytestmark = pytest.mark.asyncio


@pytest.mark.requires_pandoc
async def test_save_load_paired_md_pandoc_notebook(ipynb_py_R_jl_file, tmpdir, cm):
    if re.match(
        r".*(functional|Notebook with|flavors|invalid|305).*", ipynb_py_R_jl_file
    ):
        pytest.skip()
    tmp_ipynb = "notebook.ipynb"
    tmp_md = "notebook.md"

    cm.root_dir = str(tmpdir)

    # open ipynb, save with cm, reopen
    nb = jupytext.read(ipynb_py_R_jl_file)
    nb.metadata["jupytext"] = {"formats": "ipynb,md:pandoc"}

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))
    nb_md = await ensure_async(cm.get(tmp_md))

    compare_notebooks(nb_md["content"], nb, "md:pandoc")
    assert nb_md["content"].metadata["jupytext"]["formats"] == "ipynb,md:pandoc"


@pytest.mark.requires_quarto
async def test_save_load_paired_qmd_notebook(ipynb_py_R_jl_file, tmpdir, cm):
    if re.match(
        r".*(functional|Notebook with|plotly_graphs|flavors|complex_metadata|"
        "update83|raw_cell|_66|nteract|LaTeX|invalid|305|text_outputs|ir_notebook|jupyter|with_R_magic).*",
        ipynb_py_R_jl_file,
    ):
        pytest.skip()
    tmp_ipynb = "notebook.ipynb"
    tmp_qmd = "notebook.qmd"

    cm.root_dir = str(tmpdir)

    # open ipynb, save with cm, reopen
    nb = jupytext.read(ipynb_py_R_jl_file)
    nb.metadata["jupytext"] = {"formats": "ipynb,qmd"}

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))
    nb_md = await ensure_async(cm.get(tmp_qmd))

    compare_notebooks(nb_md["content"], nb, "qmd")
    assert nb_md["content"].metadata["jupytext"]["formats"] == "ipynb,qmd"



import pytest

import jupytext
from jupytext.compare import compare


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_read_simple_file(
    ext,
    rnb="""#' ---
#' title: Simple file
#' ---

#' Here we have some text
#' And below we have some R code

f <- function(x) {
    x + 1
    }


h <- function(y)
    y + 1
""",
):
    nb = jupytext.reads(rnb, ext)
    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "Here we have some text\n" "And below we have some R code"
    )
    assert nb.cells[2].cell_type == "code"
    compare(
        nb.cells[2].source,
        """f <- function(x) {
    x + 1
    }""",
    )
    assert nb.cells[3].cell_type == "code"
    compare(
        nb.cells[3].source,
        """h <- function(y)
    y + 1""",
    )

    rnb2 = jupytext.writes(nb, ext)
    compare(rnb2, rnb)


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_read_less_simple_file(
    ext,
    rnb="""#' ---
#' title: Less simple file
#' ---

#' Here we have some text
#' And below we have some R code

# This is a comment about function f
f <- function(x) {

    return(x+1)}


# And a comment on h
h <- function(y) {
    return(y-1)
}
""",
):
    nb = jupytext.reads(rnb, ext)

    assert len(nb.cells) == 4
    assert nb.cells[0].cell_type == "raw"
    assert nb.cells[0].source == "---\ntitle: Less simple file\n---"
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source == "Here we have some text\n" "And below we have some R code"
    )
    assert nb.cells[2].cell_type == "code"
    compare(
        nb.cells[2].source,
        """# This is a comment about function f
f <- function(x) {

    return(x+1)}""",
    )
    assert nb.cells[3].cell_type == "code"
    compare(
        nb.cells[3].source,
        """# And a comment on h
h <- function(y) {
    return(y-1)
}""",
    )

    rnb2 = jupytext.writes(nb, ext)
    compare(rnb2, rnb)


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_no_space_after_code(
    ext,
    rnb="""# -*- coding: utf-8 -*-
#' Markdown cell

f <- function(x)
{
    return(x+1)
}

#' And a new cell, and non ascii contênt
""",
):
    nb = jupytext.reads(rnb, ext)

    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "Markdown cell"
    assert nb.cells[1].cell_type == "code"
    assert (
        nb.cells[1].source
        == """f <- function(x)
{
    return(x+1)
}"""
    )
    assert nb.cells[2].cell_type == "markdown"
    assert nb.cells[2].source == "And a new cell, and non ascii contênt"

    rnb2 = jupytext.writes(nb, ext)
    compare(rnb2, rnb)


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_read_write_script(
    ext,
    rnb="""#!/usr/bin/env Rscript
# coding=utf-8
print('Hello world')
""",
):
    nb = jupytext.reads(rnb, ext)
    rnb2 = jupytext.writes(nb, ext)
    compare(rnb2, rnb)


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_escape_start_pattern(
    ext,
    rnb="""#' The code start pattern '#+' can
#' appear in code and markdown cells.

#' In markdown cells it is escaped like here:
#' #+ fig.width=12

# In code cells like this one, it is also escaped
# #+ cell_name language="python"
1 + 1
""",
):
    nb = jupytext.reads(rnb, ext)
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "markdown"
    assert nb.cells[2].cell_type == "code"
    assert (
        nb.cells[1].source
        == """In markdown cells it is escaped like here:
#+ fig.width=12"""
    )
    assert (
        nb.cells[2].source
        == """# In code cells like this one, it is also escaped
#+ cell_name language="python"
1 + 1"""
    )
    rnb2 = jupytext.writes(nb, ext)
    compare(rnb2, rnb)


@pytest.mark.parametrize("ext", [".r", ".R"])
def test_read_simple_r(
    ext,
    text="""# This is a very simple R file
# I expect to get three cells here.
#
# The first one is markdown. The two others
# are code cells

cars

plot(cars)
""",
):
    nb = jupytext.reads(text, ext)
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[1].source == "cars"
    assert nb.cells[2].source == "plot(cars)"
    text2 = jupytext.writes(nb, ext)
    compare(text2, text)



# ---
# title: Exception: Test
# jupyter:
#   jupytext:
#     cell_markers: region,endregion
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

1 + 2 + 3
